from __future__ import annotations

import logging
from numbers import Number
from typing import Dict, List

from ruamel.yaml import CommentedMap

from soda.contracts.connection import SodaException
from soda.contracts.contract import (
    Check,
    Contract,
    DuplicateCheck,
    FreshnessCheck,
    InvalidReferenceCheck,
    MissingConfigurations,
    NumericMetricCheck,
    NumericThreshold,
    Range,
    SchemaCheck,
    UserDefinedMetricSqlExpressionCheck,
    UserDefinedMetricSqlQueryCheck,
    ValidConfigurations,
    ValidValuesReferenceData,
)
from soda.contracts.impl.json_schema_verifier import JsonSchemaVerifier
from soda.contracts.impl.logs import Logs
from soda.contracts.impl.variable_resolver import VariableResolver
from soda.contracts.impl.yaml import YamlList, YamlObject, YamlParser, YamlWrapper

logger = logging.getLogger(__name__)


class ContractParser:

    __threshold_keys = [
        "must_be_greater_than",
        "must_be_greater_than_or_equal_to",
        "must_be_less_than",
        "must_be_less_than_or_equal_to",
        "must_be",
        "must_not_be",
        "must_be_between",
        "must_be_not_between",
    ]

    __validity_keys = [
        "invalid_values",
        "invalid_format",
        "invalid_sql_regex",
        "valid_values",
        "valid_format",
        "valid_sql_regex",
        "valid_min",
        "valid_max",
        "valid_length",
        "valid_min_length",
        "valid_max_length",
        "valid_values_reference_data",
    ]

    def __init__(self):
        super().__init__()
        self.logs: Logs = Logs()
        self.missing_value_configs_by_column: dict[str, MissingConfigurations] = {}
        self.valid_value_configs_by_column: dict[str, ValidConfigurations] = {}
        self.skip_schema_validation: bool = False

    def parse_contract(
        self, contract_yaml_str: str, variables: dict[str, str] | None, schedule: str | None
    ) -> Contract:
        if not isinstance(contract_yaml_str, str):
            raise SodaException(f"contract_yaml_str must be a str, but was {type(contract_yaml_str)}")

        # Resolve all the ${VARIABLES} in the contract based on either the provided
        # variables or system variables (os.environ)
        variable_resolver = VariableResolver(logs=self.logs, variables=variables)
        resolved_contract_yaml_str: str = variable_resolver.resolve(contract_yaml_str)

        yaml_parser: YamlParser = YamlParser(logs=self.logs)
        ruamel_yaml_object: object | None = yaml_parser.parse_yaml_str(yaml_str=resolved_contract_yaml_str)

        if not isinstance(ruamel_yaml_object, CommentedMap):
            raise SodaException("The contract top level must be a YAML object with keys and values")

        # Verify the contract schema on the ruamel instance object
        if ruamel_yaml_object is not None:
            json_schema_verifier: JsonSchemaVerifier = JsonSchemaVerifier(self.logs)
            json_schema_verifier.verify(ruamel_yaml_object)

        # Wrap the ruamel_yaml_object into the YamlValue (this is a better API for writing the translator)
        yaml_wrapper: YamlWrapper = YamlWrapper(logs=self.logs)
        contract_yaml_object: YamlObject = yaml_wrapper.wrap(ruamel_yaml_object)

        dataset: str | None = contract_yaml_object.read_string("dataset")
        schema_name: str | None = contract_yaml_object.read_string_opt("schema")
        sql_filter: str | None = contract_yaml_object.read_string_opt("sql_filter")

        checks: dict[str, Check] = {}

        contract_columns_yaml_list: YamlList | None = contract_yaml_object.read_yaml_list("columns")
        if contract_columns_yaml_list:
            schema_columns: Dict[str, str | None] = {}
            schema_optional_columns: List[str] = []

            schema_check_identity: str = Check.create_check_identity(
                schedule=schedule,
                dataset=dataset,
                column=None,
                check_type="schema",
                check_identity_suffix=None,
                check_location=None,
                checks=checks,
                logs=self.logs,
            )

            checks[schema_check_identity] = SchemaCheck(
                schedule=schedule,
                dataset=dataset,
                column=None,
                type="schema",
                name="Schema",
                identity=schema_check_identity,
                location=None,
                columns=schema_columns,
                optional_columns=schema_optional_columns,
            )

            contract_column_yaml_object: YamlObject
            for contract_column_yaml_object in contract_columns_yaml_list.read_yaml_objects():
                column_name: str | None = contract_column_yaml_object.read_string("name")

                if isinstance(column_name, str):
                    data_type: str | None = contract_column_yaml_object.read_string_opt("data_type")
                    schema_columns[column_name] = data_type

                    if contract_column_yaml_object.read_bool_opt("optional", default_value=False):
                        schema_optional_columns.append(column_name)

                    column_check_yaml_objects: YamlList = contract_column_yaml_object.read_yaml_list_opt("checks")

                    if column_check_yaml_objects:
                        column_check_yaml_object: YamlObject
                        for column_check_yaml_object in column_check_yaml_objects:
                            check_type: str | None = column_check_yaml_object.read_string("type")
                            check_identity_suffix: str | None = column_check_yaml_object.read_string_opt(
                                "identity_suffix"
                            )
                            if check_type is not None:
                                check_identity: str = Check.create_check_identity(
                                    schedule=schedule,
                                    dataset=dataset,
                                    column=column_name,
                                    check_type=check_type,
                                    check_identity_suffix=check_identity_suffix,
                                    check_location=column_check_yaml_object.location,
                                    checks=checks,
                                    logs=self.logs,
                                )
                                check: Check = self._parse_column_check(
                                    schedule=schedule,
                                    dataset=dataset,
                                    contract_check_id=check_identity,
                                    check_yaml_object=column_check_yaml_object,
                                    column=column_name,
                                    check_type=check_type,
                                )
                                if check:
                                    checks[check_identity] = check
                                else:
                                    logging.error(f"Could not parse check for {column_check_yaml_object.unpacked()}")

        checks_yaml_list: YamlList | None = contract_yaml_object.read_yaml_list_opt("checks")
        if checks_yaml_list:
            for check_yaml_object in checks_yaml_list:
                check_type: str | None = check_yaml_object.read_string("type")
                check_identity_suffix: str | None = check_yaml_object.read_string_opt("identity_suffix")
                if check_type is not None:
                    check_identity: str = Check.create_check_identity(
                        schedule=schedule,
                        dataset=dataset,
                        column=None,
                        check_type=check_type,
                        check_identity_suffix=check_identity_suffix,
                        check_location=check_yaml_object.location,
                        checks=checks,
                        logs=self.logs,
                    )
                    check = self._parse_dataset_check(
                        schedule=schedule,
                        dataset=dataset,
                        contract_check_id=check_identity,
                        check_yaml_object=check_yaml_object,
                        check_type=check_type,
                    )
                    if check:
                        checks[check_identity] = check

        return Contract(
            schedule=schedule,
            dataset=dataset,
            sql_filter=sql_filter,
            schema=schema_name,
            checks=list(checks.values()),
            contract_yaml_str=contract_yaml_str,
            variables=variables,
            logs=self.logs,
        )

    def _parse_column_check(
        self,
        schedule: str | None,
        dataset: str,
        column: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
    ) -> Check | None:

        if check_type in ["no_missing_values", "missing_count", "missing_percent"]:
            parse_check_function = self._parse_column_check_missing
        elif check_type in ["no_invalid_values", "invalid_count", "invalid_percent"]:
            parse_check_function = self._parse_column_check_invalid
        elif check_type in ["no_duplicate_values", "duplicate_count", "duplicate_percent"]:
            parse_check_function = self._parse_check_duplicate
        elif check_type == "metric_sql_expression":
            parse_check_function = self._parse_user_defined_metric_sql_expression_check
        elif check_type.startswith("freshness_in_"):
            parse_check_function = self._parse_freshness_check
        elif check_type in ["avg", "sum"]:
            parse_check_function = self._parse_column_check_basic_sql_function
        else:
            self.logs.error(
                f"Invalid column check type '{check_type}' (column '{column}')", location=check_yaml_object.location
            )
            return None

        return parse_check_function(
            schedule=schedule,
            dataset=dataset,
            column=column,
            contract_check_id=contract_check_id,
            check_yaml_object=check_yaml_object,
            check_type=check_type,
        )

    def _parse_column_check_missing(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ):
        name = check_yaml_object.read_string_opt("name")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        metric: str = check_type

        if check_type == "no_missing_values":
            metric = "missing_count"
            if not threshold.is_empty():
                self.logs.error(
                    f"Check type 'no_missing_values' does not allow for threshold keys must_... (column '{column}')",
                    location=check_yaml_object.location,
                )
            threshold = NumericThreshold(equal=0)
        elif threshold.is_empty():
            self.logs.error(
                (
                    f"Check type '{check_type}' requires threshold configuration "
                    f"with keys like {self.__threshold_keys} (column '{column}')"
                ),
                location=check_yaml_object.location,
            )

        missing_configurations: MissingConfigurations | None = self._parse_missing_configurations(
            check_yaml=check_yaml_object, column=column
        )

        return NumericMetricCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=missing_configurations,
            valid_configurations=None,
            threshold=threshold,
        )

    def _parse_column_check_invalid(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ):
        name = check_yaml_object.read_string_opt("name")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        metric: str = check_type

        if check_type == "no_invalid_values":
            metric = "invalid_count"
            if not threshold.is_empty():
                self.logs.error(
                    f"Check type 'no_invalid_values' does not allow for threshold "
                    f"keys must_... (column '{column}')",
                    location=check_yaml_object.location,
                )
            threshold = NumericThreshold(equal=0)
        elif threshold.is_empty():
            self.logs.error(
                (
                    f"Check type '{check_type}' requires threshold configuration "
                    f"with keys like {self.__threshold_keys} (column '{column}')"
                ),
                location=check_yaml_object.location,
            )

        missing_configurations: MissingConfigurations | None = self._parse_missing_configurations(
            check_yaml=check_yaml_object, column=column
        )

        valid_configurations: ValidConfigurations | None = self._parse_valid_configurations(
            check_yaml=check_yaml_object, column=column
        )

        if not valid_configurations:
            self.logs.error(
                message=f"Check type '{check_type}' must have a validity configuration like {self.__validity_keys} "
                f"(column '{column}')",
                location=check_yaml_object.location,
            )
            return None

        if valid_configurations.valid_values_reference_data is not None:
            if valid_configurations.has_non_reference_data_configs():
                self.logs.error(
                    message=f"Check type '{check_type}' cannot combine 'valid_values_reference_data' with other "
                    f"validity configurations (column '{column}')",
                    location=check_yaml_object.location,
                )
                return None

            return InvalidReferenceCheck(
                schedule=schedule,
                dataset=dataset,
                column=column,
                type=check_type,
                name=name,
                identity=contract_check_id,
                location=check_yaml_object.location,
                metric="invalid_count",
                check_yaml_object=check_yaml_object,
                missing_configurations=missing_configurations,
                valid_configurations=valid_configurations,
                threshold=threshold,
                valid_values_reference_data=valid_configurations.valid_values_reference_data,
            )

        return NumericMetricCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=missing_configurations,
            valid_configurations=valid_configurations,
            threshold=threshold,
        )

    def _parse_check_duplicate(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ):
        """
        Parses both the column level duplicate checks as well as the dataset level multi column duplicate checks.
        """

        name = check_yaml_object.read_string_opt("name")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        metric: str = check_type

        if check_type == "no_duplicate_values":
            metric = "duplicate_count"
            if not threshold.is_empty():
                self.logs.error(
                    f"Check type 'no_duplicate_values' does not allow for threshold keys must_... (column '{column}')",
                    location=check_yaml_object.location,
                )
            threshold = NumericThreshold(equal=0)
        elif threshold.is_empty():
            self.logs.error(
                (
                    f"Check type '{check_type}' requires threshold configuration "
                    f"with keys like {self.__threshold_keys} (column '{column}')"
                ),
                location=check_yaml_object.location,
            )

        missing_configurations: MissingConfigurations | None = None
        valid_configurations: ValidConfigurations | None = None
        columns: list[str] | None = None

        if column is None:
            # Means this is a dataset (multi-column) duplicate check
            columns = check_yaml_object.read_list_strings("columns")

        else:
            # Means this is a (single) column duplicate check
            missing_configurations = self._parse_missing_configurations(check_yaml=check_yaml_object, column=column)
            valid_configurations = self._parse_valid_configurations(check_yaml=check_yaml_object, column=column)

        return DuplicateCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=missing_configurations,
            valid_configurations=valid_configurations,
            threshold=threshold,
            columns=columns,
        )

    def _parse_column_check_basic_sql_function(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ):
        name = check_yaml_object.read_string_opt("name")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        metric: str = check_type

        return NumericMetricCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=None,
            valid_configurations=None,
            threshold=threshold,
        )

    def _parse_user_defined_metric_sql_expression_check(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ) -> Check | None:
        name = check_yaml_object.read_string_opt("name")
        metric: str = check_yaml_object.read_string("metric")
        sql_expression: str = check_yaml_object.read_string("sql_expression")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        if not threshold:
            self.logs.error("No threshold defined for metric_sql_expression check", location=check_yaml_object.location)

        return UserDefinedMetricSqlExpressionCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=None,
            valid_configurations=None,
            threshold=threshold,
            metric_sql_expression=sql_expression,
        )

    def _parse_freshness_check(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ) -> Check | None:
        name = check_yaml_object.read_string_opt("name")

        freshness_check_types = [
            "freshness_in_days",
            "freshness_in_hours",
            "freshness_in_minutes",
        ]
        if check_type not in freshness_check_types:
            self.logs.error(f"Invalid freshness check type: {check_type}: Expected one of {freshness_check_types}")
            return None

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)
        if not threshold:
            self.logs.error("No threshold defined for freshness check", location=check_yaml_object.location)
        elif (
            threshold.not_between is not None
            or threshold.between is not None
            or threshold.equal is not None
            or threshold.not_equal is not None
            or threshold.greater_than is not None
            or threshold.greater_than is not None
            or threshold.less_than_or_equal is not None
        ):
            self.logs.error("Invalid freshness threshold. Use must_be_less_than", location=check_yaml_object.location)

        return FreshnessCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            threshold=threshold,
        )

    def _parse_missing_configurations(self, check_yaml: YamlObject, column: str) -> MissingConfigurations | None:
        missing_values_yaml_list: YamlList | None = check_yaml.read_yaml_list_opt(f"missing_values")
        missing_values: list | None = missing_values_yaml_list.unpacked() if missing_values_yaml_list else None
        missing_sql_regex: str | None = check_yaml.read_string_opt(f"missing_sql_regex")

        if all(v is None for v in [missing_values, missing_sql_regex]):
            return self.missing_value_configs_by_column.get(column)

        else:
            missing_configurations = MissingConfigurations(
                missing_values=missing_values, missing_sql_regex=missing_sql_regex
            )

            # If a missing config is specified, do a complete overwrite.
            # Overwriting the missing configs gives more control to the contract author over merging the missing configs.
            self.missing_value_configs_by_column[column] = missing_configurations

            return missing_configurations

    def _parse_valid_configurations(self, check_yaml: YamlObject, column: str) -> ValidConfigurations | None:
        invalid_values_yaml_list: YamlList | None = check_yaml.read_yaml_list_opt(f"invalid_values")
        invalid_values: list | None = invalid_values_yaml_list.unpacked() if invalid_values_yaml_list else None
        invalid_format: str | None = check_yaml.read_string_opt(f"invalid_format")
        invalid_sql_regex: str | None = check_yaml.read_string_opt(f"invalid_sql_regex")

        valid_values_yaml_list: YamlList | None = check_yaml.read_yaml_list_opt(f"valid_values")
        valid_values: list | None = valid_values_yaml_list.unpacked() if valid_values_yaml_list else None

        valid_format: str | None = check_yaml.read_string_opt(f"valid_format")
        valid_sql_regex: str | None = check_yaml.read_string_opt(f"valid_sql_regex")

        valid_min: Number | None = check_yaml.read_number_opt(f"valid_min")
        valid_max: Number | None = check_yaml.read_number_opt(f"valid_max")

        valid_length: int | None = check_yaml.read_number_opt(f"valid_length")
        valid_min_length: int | None = check_yaml.read_number_opt(f"valid_min_length")
        valid_max_length: int | None = check_yaml.read_number_opt(f"valid_max_length")

        valid_values_reference_data: ValidValuesReferenceData | None = None
        valid_values_reference_data_yaml_object: YamlObject | None = check_yaml.read_yaml_object_opt(
            f"valid_values_reference_data"
        )
        if valid_values_reference_data_yaml_object:
            ref_dataset = valid_values_reference_data_yaml_object.read_string("dataset")
            ref_column = valid_values_reference_data_yaml_object.read_string("column")
            valid_values_reference_data = ValidValuesReferenceData(dataset=ref_dataset, column=ref_column)

        if all(
            v is None
            for v in [
                invalid_values,
                invalid_format,
                invalid_sql_regex,
                valid_values,
                valid_format,
                valid_sql_regex,
                valid_min,
                valid_max,
                valid_length,
                valid_min_length,
                valid_max_length,
                valid_values_reference_data,
            ]
        ):
            return self.valid_value_configs_by_column.get(column)
        else:
            valid_configurations = ValidConfigurations(
                invalid_values=invalid_values,
                invalid_format=invalid_format,
                invalid_sql_regex=invalid_sql_regex,
                valid_values=valid_values,
                valid_format=valid_format,
                valid_sql_regex=valid_sql_regex,
                valid_min=valid_min,
                valid_max=valid_max,
                valid_length=valid_length,
                valid_min_length=valid_min_length,
                valid_max_length=valid_max_length,
                valid_values_reference_data=valid_values_reference_data,
            )

            # If a valid config is specified, do a complete overwrite.
            # Overwriting the valid configs gives more control to the contract author over merging the missing configs.
            self.valid_value_configs_by_column[column] = valid_configurations

            return valid_configurations

    def _parse_numeric_threshold(self, check_yaml_object: YamlObject) -> NumericThreshold | None:
        numeric_threshold: NumericThreshold = NumericThreshold(
            greater_than=check_yaml_object.read_number_opt("must_be_greater_than"),
            greater_than_or_equal=check_yaml_object.read_number_opt("must_be_greater_than_or_equal_to"),
            less_than=check_yaml_object.read_number_opt("must_be_less_than"),
            less_than_or_equal=check_yaml_object.read_number_opt("must_be_less_than_or_equal_to"),
            equal=check_yaml_object.read_number_opt("must_be"),
            not_equal=check_yaml_object.read_number_opt("must_not_be"),
            between=self._parse_range(check_yaml_object, "must_be_between"),
            not_between=self._parse_range(check_yaml_object, "must_be_not_between"),
        )

        for key in check_yaml_object:
            if key.startswith("must_") and key not in self.__threshold_keys:
                self.logs.error(f"Invalid threshold '{key}'. Must be in '{self.__threshold_keys}'.")

        return numeric_threshold

    def _parse_range(self, check_yaml_object: YamlObject, range_key: str) -> Range | None:
        range_yaml_list: YamlList | None = check_yaml_object.read_yaml_list_opt(range_key)
        if isinstance(range_yaml_list, YamlList):
            range_values = range_yaml_list.unpacked()
            if all(isinstance(range_value, Number) for range_value in range_values) and len(range_values) == 2:
                return Range(lower_bound=range_values[0], upper_bound=range_values[1])
            else:
                self.logs.error("range expects a list of 2 numbers", location=range_yaml_list.location)

    def _parse_dataset_check(
        self, schedule: str | None, dataset: str, contract_check_id: str, check_yaml_object: YamlObject, check_type: str
    ) -> Check | None:

        if check_type in ["rows_exist", "row_count"]:
            check_parse_function = self._parse_dataset_row_count
        elif check_type in ["no_duplicate_values", "duplicate_count", "duplicate_percent"]:
            check_parse_function = self._parse_check_duplicate
        elif check_type == "metric_sql_expression":
            check_parse_function = self._parse_user_defined_metric_sql_expression_check
        elif check_type == "metric_sql_query":
            check_parse_function = self._parse_user_defined_metric_sql_query_check
        elif check_type == "failed_rows_sql_query":
            check_parse_function = self._parse_user_defined_failed_rows_query_check
        else:
            self.logs.error(f"Invalid dataset check type '{check_type}'", location=check_yaml_object.location)
            return None

        return check_parse_function(
            schedule=schedule,
            dataset=dataset,
            column=None,
            contract_check_id=contract_check_id,
            check_yaml_object=check_yaml_object,
            check_type=check_type,
        )

    def _parse_dataset_row_count(
        self,
        schedule: str | None,
        dataset: str,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
        column: str | None,
    ) -> NumericMetricCheck:
        name = check_yaml_object.read_string_opt("name")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        metric: str = check_type

        if check_type == "rows_exist":
            metric = "row_count"
            if not threshold.is_empty():
                self.logs.error(
                    f"Check type 'rows_exist' does not allow for threshold keys must_... (column '{column}')",
                    location=check_yaml_object.location,
                )
            threshold = NumericThreshold(greater_than=0)
        elif threshold.is_empty():
            self.logs.error(
                (
                    f"Check type '{check_type}' requires threshold configuration "
                    f"with keys like {self.__threshold_keys} (column '{column}')"
                ),
                location=check_yaml_object.location,
            )

        return NumericMetricCheck(
            schedule=schedule,
            dataset=dataset,
            column=column,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=None,
            valid_configurations=None,
            threshold=threshold,
        )

    def _parse_user_defined_metric_sql_query_check(
        self,
        schedule: str | None,
        dataset: str,
        column: None,
        contract_check_id: str,
        check_yaml_object: YamlObject,
        check_type: str,
    ) -> Check | None:

        name = check_yaml_object.read_string_opt("name")
        metric: str = check_yaml_object.read_string("metric")
        metric_sql_query: str = check_yaml_object.read_string("sql_query")

        threshold: NumericThreshold = self._parse_numeric_threshold(check_yaml_object=check_yaml_object)

        return UserDefinedMetricSqlQueryCheck(
            schedule=schedule,
            dataset=dataset,
            column=None,
            type=check_type,
            name=name,
            identity=contract_check_id,
            location=check_yaml_object.location,
            check_yaml_object=check_yaml_object,
            metric=metric,
            missing_configurations=None,
            valid_configurations=None,
            threshold=threshold,
            metric_sql_query=metric_sql_query,
        )



from __future__ import annotations

import dataclasses
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from numbers import Number
from textwrap import indent
from typing import Dict, List

from soda.cloud.soda_cloud import SodaCloud
from soda.common import logs as soda_core_logs
from soda.scan import Scan
from soda.scan import logger as scan_logger

from soda.contracts import soda_cloud as contract_soda_cloud
from soda.contracts.connection import SodaException
from soda.contracts.impl.logs import Location, Log, LogLevel, Logs
from soda.contracts.impl.yaml import YamlObject, YamlWriter

logger = logging.getLogger(__name__)


class Contract:

    @classmethod
    def from_yaml_str(
        cls, contract_yaml_str: str, variables: dict[str, str] | None = None, schedule: str | None = None
    ) -> Contract:
        """
        Build a contract from a YAML string
        """
        from soda.contracts.impl.contract_parser import ContractParser

        contract_parser: ContractParser = ContractParser()
        return contract_parser.parse_contract(
            contract_yaml_str=contract_yaml_str, variables=variables, schedule=schedule
        )

    @classmethod
    def from_yaml_file(
        cls, file_path: str, variables: dict[str, str] | None = None, schedule: str | None = None
    ) -> Contract:
        """
        Build a contract from a YAML file.
        Raises OSError in case the file_path cannot be opened like e.g.
        FileNotFoundError or PermissionError
        """
        with open(file_path) as f:
            contract_yaml_str = f.read()
            return cls.from_yaml_str(contract_yaml_str=contract_yaml_str, variables=variables, schedule=schedule)

    def __init__(
        self,
        schedule: str | None,
        dataset: str,
        sql_filter: str | None,
        schema: str | None,
        checks: List[Check],
        contract_yaml_str: str,
        variables: dict[str, str] | None,
        logs: Logs,
    ):
        """
        Consider using Contract.from_yaml_str(contract_yaml_str) instead as that is more stable API.
        """
        self.schedule: str | None = schedule
        self.dataset: str = dataset
        self.sql_filter: str | None = sql_filter
        self.schema: str | None = schema
        self.checks: List[Check] = checks
        self.contract_yaml_str: str = contract_yaml_str
        self.variables: dict[str, str] | None = variables
        # The initial logs will contain the logs of contract parser.  If there are error logs, these error logs
        # will cause a SodaException to be raised at the end of the Contract.verify method
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs
        self.sodacl_yaml_str: str | None = None

    def verify(self, connection: Connection, soda_cloud: contract_soda_cloud.SodaCloud | None = None) -> ContractResult:
        """
        Verifies if the data in the dataset matches the contract.
        """

        scan = Scan()

        scan_logs = soda_core_logs.Logs(logger=scan_logger)
        scan_logs.verbose = True

        sodacl_yaml_str: str | None = None
        try:
            sodacl_yaml_str = self.generate_sodacl_yaml_str(self.logs)
            logger.debug(sodacl_yaml_str)

            if sodacl_yaml_str and hasattr(connection, "data_source"):
                scan._logs = scan_logs

                # This assumes the connection is a DataSourceConnection
                data_source = connection.data_source
                # Execute the contract SodaCL in a scan
                scan.set_data_source_name(data_source.data_source_name)
                scan_definition_name = (
                    f"dataset://{connection.name}/{self.schema}/{self.dataset}"
                    if self.schema
                    else f"dataset://{connection.name}/{self.dataset}"
                )
                scan._data_source_manager.data_sources[data_source.data_source_name] = data_source

                if soda_cloud:
                    scan.set_scan_definition_name(scan_definition_name)
                    scan._configuration.soda_cloud = SodaCloud(
                        host=soda_cloud.host,
                        api_key_id=soda_cloud.api_key_id,
                        api_key_secret=soda_cloud.api_key_secret,
                        token=soda_cloud.token,
                        port=soda_cloud.port,
                        logs=scan_logs,
                        scheme=soda_cloud.scheme,
                    )

                if self.variables:
                    scan.add_variables(self.variables)

                # noinspection PyProtectedMember
                scan.add_sodacl_yaml_str(sodacl_yaml_str)
                scan.execute()

        except Exception as e:
            self.logs.error(f"Data contract verification error: {e}", exception=e)

        if soda_cloud:
            # If SodaCloud is configured, the logs are copied into the contract result and
            # at the end of this method, a SodaException is raised if there are error logs.
            self.logs.logs.extend(soda_cloud.logs.logs)
        if connection:
            # The connection logs are copied into the contract result and at the end of this
            # method, a SodaException is raised if there are error logs.
            self.logs.logs.extend(connection.logs.logs)
        # The scan warning and error logs are copied into self.logs and at the end of this
        # method, a SodaException is raised if there are error logs.
        self.append_scan_warning_and_error_logs(scan_logs)

        contract_result: ContractResult = ContractResult(
            contract=self, sodacl_yaml_str=sodacl_yaml_str, logs=self.logs, scan=scan
        )
        if contract_result.failed():
            raise SodaException(contract_result=contract_result)

        return contract_result

    def append_scan_warning_and_error_logs(self, scan_logs: soda_core_logs.Logs) -> None:
        level_map = {
            soda_core_logs.LogLevel.ERROR: LogLevel.ERROR,
            soda_core_logs.LogLevel.WARNING: LogLevel.WARNING,
            soda_core_logs.LogLevel.INFO: LogLevel.INFO,
            soda_core_logs.LogLevel.DEBUG: LogLevel.DEBUG,
        }
        for scan_log in scan_logs.logs:
            if scan_log.level in [soda_core_logs.LogLevel.ERROR, soda_core_logs.LogLevel.WARNING]:
                contracts_location: Location = (
                    Location(line=scan_log.location.line, column=scan_log.location.col)
                    if scan_log.location is not None
                    else None
                )
                contracts_level: LogLevel = level_map[scan_log.level]
                self.logs._log(
                    Log(
                        level=contracts_level,
                        message=f"SodaCL: {scan_log.message}",
                        location=contracts_location,
                        exception=scan_log.exception,
                    )
                )

    def generate_sodacl_yaml_str(self, logs: Logs) -> str:
        # Serialize the SodaCL YAML object to a YAML string
        sodacl_checks: list = []
        sodacl_yaml_object: dict = (
            {
                f"filter {self.dataset} [filter]": {"where": self.sql_filter},
                f"checks for {self.dataset} [filter]": sodacl_checks,
            }
            if self.sql_filter
            else {f"checks for {self.dataset}": sodacl_checks}
        )

        for check in self.checks:
            sodacl_check = check.to_sodacl_check()
            if sodacl_check is not None:
                sodacl_checks.append(sodacl_check)
        yaml_writer: YamlWriter = YamlWriter(logs)
        return yaml_writer.write_to_yaml_str(sodacl_yaml_object)


@dataclass
class ContractResult:
    """
    This is the immutable data structure containing all the results from a single contract verification.
    This includes any potential execution errors as well as the results of all the checks performed.
    """

    contract: Contract
    sodacl_yaml_str: str | None
    # self.logs combines all the logs of the contract verification with the logs of the Connection parsing,
    # connection usage, SodaCloud parsing and usage (if used) and contract parsing.
    # At the end of the verify method a SodaException is raised if there are any error logs or check failures.
    # See also adr/03_exceptions_vs_error_logs.md
    logs: Logs
    check_results: List[CheckResult]

    def __init__(self, contract: Contract, sodacl_yaml_str: str | None, logs: Logs, scan: Scan):
        self.contract = contract
        self.sodacl_yaml_str = sodacl_yaml_str
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = Logs(logs)
        self.check_results: List[CheckResult] = []

        contract_checks_by_id: dict[str, Check] = {check.identity: check for check in contract.checks}

        schema_check: SchemaCheck | None = next((c for c in contract.checks if isinstance(c, SchemaCheck)), None)

        scan_metrics_by_id: dict[str, dict] = {
            scan_metric["identity"]: scan_metric for scan_metric in scan.scan_results.get("metrics", [])
        }

        scan_checks = scan.scan_results.get("checks")
        if isinstance(scan_checks, list):
            for scan_check in scan_checks:
                contract_check: Check | None = None
                if scan_check.get("name") == "Schema Check" and scan_check.get("type") == "generic":
                    contract_check = schema_check
                else:
                    contract_check_id = scan_check.get("contract_check_id")
                    if isinstance(contract_check_id, str):
                        contract_check = contract_checks_by_id[contract_check_id]

                assert contract_check is not None, "Contract scan check matching failed :("

                scan_check_metric_ids = scan_check.get("metrics")
                scan_check_metrics = [
                    scan_metrics_by_id.get(check_metric_id) for check_metric_id in scan_check_metric_ids
                ]
                scan_check_metrics_by_name = {
                    scan_check_metric.get("metricName"): scan_check_metric for scan_check_metric in scan_check_metrics
                }
                check_result = contract_check.create_check_result(
                    scan_check=scan_check, scan_check_metrics_by_name=scan_check_metrics_by_name, scan=scan
                )
                self.check_results.append(check_result)

    def failed(self) -> bool:
        return self.has_execution_errors() or self.has_check_failures()

    def passed(self) -> bool:
        return not self.failed()

    def has_execution_errors(self):
        return self.logs.has_errors()

    def has_check_failures(self):
        return any(check.outcome == CheckOutcome.FAIL for check in self.check_results)

    def __str__(self) -> str:
        error_texts_list: List[str] = [str(error) for error in self.logs.get_errors()]

        check_failure_message_list: list[str] = []
        for check_result in self.check_results:
            if check_result.outcome == CheckOutcome.FAIL:
                result_str_lines = check_result.get_contract_result_str_lines()
                check_failure_message_list.extend(result_str_lines)

        if not error_texts_list and not check_failure_message_list:
            return "All is good. No checks failed. No contract execution errors."

        errors_summary_text = f"{len(error_texts_list)} execution error"
        if len(error_texts_list) != 1:
            errors_summary_text = f"{errors_summary_text}s"

        checks_summary_text = f"{len(check_failure_message_list)} check failure"
        if len(check_failure_message_list) != 1:
            checks_summary_text = f"{checks_summary_text}s"

        parts = [f"{checks_summary_text} and {errors_summary_text}"]
        if error_texts_list:
            error_lines_text: str = indent("\n".join(error_texts_list), "  ")
            parts.append(f"Errors: \n{error_lines_text}")

        if check_failure_message_list:
            parts.append("\n".join(check_failure_message_list))

        return "\n".join(parts)


@dataclass
class Check(ABC):

    schedule: str | None
    dataset: str
    column: str | None
    type: str

    # User defined name as in the contract.  None if not specified in the contract.
    name: str | None

    # Check identifier used to correlate the sodacl check results with this contract check object when parsing
    # scan results.  Also used as correlation id in Soda Cloud to match subsequent results for the same check.
    # Composite key created from schedule, dataset, column, type and identity_suffix.
    identity: str | None

    location: Location | None

    @abstractmethod
    def to_sodacl_check(self) -> str | dict | None:
        pass

    @abstractmethod
    def create_check_result(
        self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan
    ) -> CheckResult:
        pass

    @classmethod
    def create_check_identity(
        cls,
        schedule: str | None,
        dataset: str,
        column: str | None,
        check_type: str,
        check_identity_suffix: str | None,
        check_location: Location | None,
        checks: dict[str, Check],
        logs: Logs,
    ) -> str:
        opt_schedule_part = f"//{schedule}" if schedule else ""
        opt_column_part = f"/{column}" if column else ""
        opt_check_identity_suffix_part = f"/{check_identity_suffix}" if check_identity_suffix else ""
        check_identity = f"{opt_schedule_part}/{dataset}{opt_column_part}/{check_type}{opt_check_identity_suffix_part}"

        other_check: Check = checks.get(check_identity)
        if other_check:
            logs.error(f"Duplicate check identity '{check_identity}': {other_check.location} and {check_location}")
            suffix_index: int = 2
            while check_identity in checks:
                opt_check_identity_suffix_part = (
                    f"/{check_identity_suffix}_{suffix_index}" if check_identity_suffix else f"/{suffix_index}"
                )
                check_identity = (
                    f"{opt_schedule_part}/{dataset}{opt_column_part}/{check_type}{opt_check_identity_suffix_part}"
                )
                suffix_index = suffix_index + 1
        return check_identity


@dataclass
class CheckResult:
    check: Check
    outcome: CheckOutcome

    def __str__(self) -> str:
        return "\n".join(self.get_contract_result_str_lines())

    @abstractmethod
    def get_contract_result_str_lines(self) -> list[str]:
        """
        Provides the summary for the contract result logs, as well as the __str__ impl of this check result.
        Method implementations can use self._get_outcome_line(self)
        """

    def get_outcome_and_name_line(self) -> str:
        name_str: str = f" [{self.check.name}]" if self.check.name else ""
        return f"Check {self.get_outcome_str()}{name_str}"

    def get_outcome_str(self) -> str:
        if self.outcome == CheckOutcome.FAIL:
            return "FAILED"
        if self.outcome == CheckOutcome.PASS:
            return "passed"
        return "unverified"


@dataclass
class SchemaCheck(Check):

    columns: dict[str, str | None]
    optional_columns: list[str]

    def to_sodacl_check(self) -> str | dict | None:
        schema_fail_dict = {"when mismatching columns": self.columns}
        if self.optional_columns:
            schema_fail_dict["with optional columns"] = self.optional_columns
        return {"schema": {"fail": schema_fail_dict}}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_measured_schema: list[dict] = scan_check_metrics_by_name.get("schema").get("value")
        measured_schema = {c.get("columnName"): c.get("sourceDataType") for c in scan_measured_schema}

        diagnostics = scan_check.get("diagnostics", {})

        columns_not_allowed_and_present: list[str] = diagnostics.get("present_column_names", [])
        columns_required_and_not_present: list[str] = diagnostics.get("missing_column_names", [])

        columns_having_wrong_type: list[DataTypeMismatch] = []
        scan_column_type_mismatches = diagnostics.get("column_type_mismatches", {})
        if scan_column_type_mismatches:
            for column_name, column_type_mismatch in scan_column_type_mismatches.items():
                expected_type = column_type_mismatch.get("expected_type")
                actual_type = column_type_mismatch.get("actual_type")
                columns_having_wrong_type.append(
                    DataTypeMismatch(column=column_name, expected_data_type=expected_type, actual_data_type=actual_type)
                )

        return SchemaCheckResult(
            check=self,
            outcome=CheckOutcome.from_scan_check(scan_check),
            measured_schema=measured_schema,
            columns_not_allowed_and_present=columns_not_allowed_and_present,
            columns_required_and_not_present=columns_required_and_not_present,
            columns_having_wrong_type=columns_having_wrong_type,
        )


@dataclass
class SchemaCheckResult(CheckResult):

    measured_schema: Dict[str, str]
    columns_not_allowed_and_present: list[str] | None
    columns_required_and_not_present: list[str] | None
    columns_having_wrong_type: list[DataTypeMismatch] | None

    def get_contract_result_str_lines(self) -> list[str]:
        schema_check: SchemaCheck = self.check
        expected_schema: str = ",".join(
            [
                f"{c.get('name')}{c.get('optional')}{c.get('type')}"
                for c in [
                    {
                        "name": column_name,
                        "optional": "(optional)" if column_name in schema_check.optional_columns else "",
                        "type": f"={data_type}" if data_type else "",
                    }
                    for column_name, data_type in schema_check.columns.items()
                ]
            ]
        )

        lines: list[str] = [
            f"Schema check {self.get_outcome_str()}",
            f"  Expected schema: {expected_schema}",
            f"  Actual schema: {self.measured_schema}",
        ]
        lines.extend(
            [f"  Column '{column}' was present and not allowed" for column in self.columns_not_allowed_and_present]
        )
        lines.extend([f"  Column '{column}' was missing" for column in self.columns_required_and_not_present])
        lines.extend(
            [
                (
                    f"  Column '{data_type_mismatch.column}': Expected type '{data_type_mismatch.expected_data_type}', "
                    f"but was '{data_type_mismatch.actual_data_type}'"
                )
                for data_type_mismatch in self.columns_having_wrong_type
            ]
        )
        return lines


@dataclass
class NumericMetricCheck(Check):

    metric: str
    check_yaml_object: YamlObject
    missing_configurations: MissingConfigurations | None
    valid_configurations: ValidConfigurations | None
    threshold: NumericThreshold | None

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_line = self.get_sodacl_check_line()

        sodacl_check_configs = {"contract check id": self.identity}

        if self.name:
            sodacl_check_configs["name"] = self.name

        if self.valid_configurations:
            sodacl_check_configs.update(self.valid_configurations.to_sodacl_check_configs_dict())
        if self.missing_configurations:
            sodacl_check_configs.update(self.missing_configurations.to_sodacl_check_configs_dict())

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict: dict
        if "(" in self.metric:
            scan_metric_name = self.metric[: self.metric.index("(")]
            scan_metric_dict = scan_check_metrics_by_name.get(scan_metric_name, None)
        else:
            scan_metric_dict = scan_check_metrics_by_name.get(self.metric, None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None
        return NumericMetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )

    def get_sodacl_check_line(self) -> str:
        sodacl_metric = self.get_sodacl_metric()
        sodacl_threshold: str = self.threshold.get_sodacl_threshold() if self.threshold else ""
        return f"{sodacl_metric} {sodacl_threshold}"

    def get_sodacl_metric(self) -> str:
        return f"{self.metric}({self.column})" if self.column else self.metric

    def get_sodacl_threshold(self) -> str:
        return self.threshold.get_sodacl_threshold() if self.threshold else "?"

    def get_metric_str(self) -> str:
        return self.get_sodacl_metric()

    def get_expected_str(self) -> str:
        return f"{self.get_metric_str()} {self.get_sodacl_threshold()}"


@dataclass
class NumericMetricCheckResult(CheckResult):
    metric_value: Number

    def get_contract_result_str_lines(self) -> list[str]:
        return [
            self.get_outcome_and_name_line(),
            f"  Expected {self.check.get_expected_str()}",
            f"  Actual {self.check.get_metric_str() } was {self.metric_value}",
        ]


@dataclass
class DuplicateCheck(NumericMetricCheck):
    columns: list[str]

    def get_sodacl_metric(self) -> str:
        column_str = self.column if self.column else ", ".join(self.columns)
        return f"{self.metric}({column_str})"


@dataclass
class InvalidReferenceCheck(NumericMetricCheck):

    valid_values_reference_data: ValidValuesReferenceData

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = {"contract check id": self.identity}

        if self.name:
            sodacl_check_configs["name"] = self.name

        if self.valid_configurations:
            sodacl_check_configs.update(self.valid_configurations.to_sodacl_check_configs_dict())
        if self.missing_configurations:
            sodacl_check_configs.update(self.missing_configurations.to_sodacl_check_configs_dict())

        sodacl_check_line: str = (
            f"values in ({self.column}) must exist in {self.valid_values_reference_data.dataset} ({self.valid_values_reference_data.column})"
        )

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict = scan_check_metrics_by_name.get("reference", {})
        value: Number = scan_metric_dict.get("value")
        return NumericMetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=value
        )


@dataclass
class FreshnessCheck(Check):

    column: str | None
    check_yaml_object: YamlObject
    threshold: NumericThreshold | None

    def get_definition_line(self) -> str:
        return f"freshness({self.column}) {self.threshold.get_sodacl_threshold()}{self.get_sodacl_time_unit()}"

    def get_sodacl_time_unit(self) -> str:
        sodacl_time_unit_by_check_type = {
            "freshness_in_days": "d",
            "freshness_in_hours": "h",
            "freshness_in_minutes": "m",
        }
        return sodacl_time_unit_by_check_type.get(self.type)

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = {
            "contract check id": self.identity,
        }
        if self.name:
            sodacl_check_configs["name"] = self.name

        sodacl_check_line: str = self.get_definition_line()
        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        diagnostics: dict = scan_check["diagnostics"]
        freshness = diagnostics["freshness"]
        freshness_column_max_value = diagnostics["maxColumnTimestamp"]
        freshness_column_max_value_utc = diagnostics["maxColumnTimestampUtc"]
        now = diagnostics["nowTimestamp"]
        now_utc = diagnostics["nowTimestampUtc"]

        return FreshnessCheckResult(
            check=self,
            outcome=CheckOutcome.from_scan_check(scan_check),
            freshness=freshness,
            freshness_column_max_value=freshness_column_max_value,
            freshness_column_max_value_utc=freshness_column_max_value_utc,
            now=now,
            now_utc=now_utc,
        )


@dataclass
class FreshnessCheckResult(CheckResult):
    freshness: str
    freshness_column_max_value: str
    freshness_column_max_value_utc: str
    now: str
    now_utc: str

    def get_contract_result_str_lines(self) -> list[str]:
        assert isinstance(self.check, FreshnessCheck)
        return [
            self.get_outcome_and_name_line(),
            f"  Expected {self.check.get_definition_line()}",
            f"  Actual freshness({self.check.column}) was {self.freshness}",
            f"  Max value in column was ...... {self.freshness_column_max_value}",
            f"  Max value in column in UTC was {self.freshness_column_max_value_utc}",
            f"  Now was ...................... {self.now}",
            f"  Now in UTC was ............... {self.now_utc}",
        ]


@dataclass
class UserDefinedMetricSqlExpressionCheck(NumericMetricCheck):

    metric_sql_expression: str

    def to_sodacl_check(self) -> str | dict | None:

        sodacl_check_configs = {
            "contract check id": self.identity,
            f"{self.metric} expression": self.metric_sql_expression,
        }
        if self.name:
            sodacl_check_configs["name"] = self.name

        sodacl_checkline_threshold = self.threshold.get_sodacl_threshold()
        sodacl_check_line = f"{self.get_sodacl_metric()} {sodacl_checkline_threshold}"

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict: dict = scan_check_metrics_by_name.get(self.metric, None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None
        return NumericMetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )


@dataclass
class UserDefinedMetricSqlQueryCheck(NumericMetricCheck):

    metric_sql_query: str

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = {
            "contract check id": self.identity,
            f"{self.metric} query": self.metric_sql_query,
        }
        if self.name:
            sodacl_check_configs["name"] = self.name

        sodacl_check_line: str = self.get_sodacl_check_line()

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict: dict = scan_check_metrics_by_name.get(self.get_sodacl_check_line(), None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None

        return NumericMetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )


class CheckOutcome(Enum):
    PASS = "pass"
    FAIL = "fail"
    UNKNOWN = "unknown"

    @classmethod
    def from_scan_check(cls, scan_check: Dict[str, object]) -> CheckOutcome:
        scan_check_outcome = scan_check.get("outcome")
        if scan_check_outcome == "pass":
            return CheckOutcome.PASS
        elif scan_check_outcome == "fail":
            return CheckOutcome.FAIL
        return CheckOutcome.UNKNOWN


@dataclass
class DataTypeMismatch:
    column: str
    expected_data_type: str
    actual_data_type: str


def dataclass_object_to_sodacl_dict(dataclass_object: object) -> dict:
    def translate_to_sodacl_key(key: str) -> str:
        if "sql_" in key:
            key = key.replace("sql_", "")
        return key.replace("_", " ")

    dict_factory = lambda x: {translate_to_sodacl_key(k): v for (k, v) in x if v is not None}
    return dataclasses.asdict(dataclass_object, dict_factory=dict_factory)


@dataclass
class MissingConfigurations:
    missing_values: list[str] | list[Number] | None
    missing_sql_regex: str | None

    def to_sodacl_check_configs_dict(self) -> dict:
        return dataclass_object_to_sodacl_dict(self)


@dataclass
class ValidConfigurations:
    invalid_values: list[str] | list[Number] | None
    invalid_format: str | None
    invalid_sql_regex: str | None
    valid_values: list[str] | list[Number] | None
    valid_format: str | None
    valid_sql_regex: str | None
    valid_min: Number | None
    valid_max: Number | None
    valid_length: int | None
    valid_min_length: int | None
    valid_max_length: int | None
    valid_values_reference_data: ValidValuesReferenceData | None

    def to_sodacl_check_configs_dict(self) -> dict:
        sodacl_check_configs_dict = dataclass_object_to_sodacl_dict(self)
        sodacl_check_configs_dict.pop("valid values reference data", None)
        return sodacl_check_configs_dict

    def has_non_reference_data_configs(self) -> bool:
        return (
            self.invalid_values is not None
            or self.invalid_format is not None
            or self.invalid_sql_regex is not None
            or self.valid_values is not None
            or self.valid_format is not None
            or self.valid_sql_regex is not None
            or self.valid_min is not None
            or self.valid_max is not None
            or self.valid_length is not None
            or self.valid_min_length is not None
            or self.valid_max_length is not None
        )


@dataclass
class ValidValuesReferenceData:
    dataset: str
    column: str


@dataclass
class NumericThreshold:
    """
    The threshold is exceeded when any of the member field conditions is True.
    To be interpreted as a check fails when the metric value is ...greater_than or ...less_than etc...
    """

    greater_than: Number | None = None
    greater_than_or_equal: Number | None = None
    less_than: Number | None = None
    less_than_or_equal: Number | None = None
    equal: Number | None = None
    not_equal: Number | None = None
    between: Range | None = None
    not_between: Range | None = None

    def get_sodacl_threshold(self) -> str:
        greater_bound: Number | None = (
            self.greater_than if self.greater_than is not None else self.greater_than_or_equal
        )
        less_bound: Number | None = self.less_than if self.less_than is not None else self.less_than_or_equal
        if greater_bound is not None and less_bound is not None:
            if greater_bound > less_bound:
                return self.sodacl_threshold(
                    is_not_between=True,
                    lower_bound=less_bound,
                    lower_bound_included=self.less_than is not None,
                    upper_bound=greater_bound,
                    upper_bound_included=self.greater_than is not None,
                )
            else:
                return self.sodacl_threshold(
                    is_not_between=False,
                    lower_bound=greater_bound,
                    lower_bound_included=self.greater_than_or_equal is not None,
                    upper_bound=less_bound,
                    upper_bound_included=self.less_than_or_equal is not None,
                )
        elif isinstance(self.between, Range):
            return self.sodacl_threshold(
                is_not_between=False,
                lower_bound=self.between.lower_bound,
                lower_bound_included=True,
                upper_bound=self.between.upper_bound,
                upper_bound_included=True,
            )
        elif isinstance(self.not_between, Range):
            return self.sodacl_threshold(
                is_not_between=True,
                lower_bound=self.not_between.lower_bound,
                lower_bound_included=True,
                upper_bound=self.not_between.upper_bound,
                upper_bound_included=True,
            )
        elif self.greater_than is not None:
            return f"> {self.greater_than}"
        elif self.greater_than_or_equal is not None:
            return f">= {self.greater_than_or_equal}"
        elif self.less_than is not None:
            return f"< {self.less_than}"
        elif self.less_than_or_equal is not None:
            return f"<= {self.less_than_or_equal}"
        elif self.equal is not None:
            return f"= {self.equal}"
        elif self.not_equal is not None:
            return f"!= {self.not_equal}"

    @classmethod
    def sodacl_threshold(
        cls,
        is_not_between: bool,
        lower_bound: Number,
        lower_bound_included: bool,
        upper_bound: Number,
        upper_bound_included: bool,
    ) -> str:
        optional_not = "not " if is_not_between else ""
        lower_bound_bracket = "" if lower_bound_included else "("
        upper_bound_bracket = "" if upper_bound_included else ")"
        return f"{optional_not}between {lower_bound_bracket}{lower_bound} and {upper_bound}{upper_bound_bracket}"

    def is_empty(self) -> bool:
        return (
            self.greater_than is None
            and self.greater_than_or_equal is None
            and self.less_than is None
            and self.less_than_or_equal is None
            and self.equal is None
            and self.not_equal is None
            and self.between is None
            and self.not_between is None
        )


@dataclass
class Range:
    """
    Boundary values are inclusive
    """

    lower_bound: Number | None
    upper_bound: Number | None



from __future__ import annotations

import os

from ruamel.yaml import YAML
from yaml import MarkedYAMLError

from soda.contracts.connection import SodaException
from soda.contracts.impl.logs import Logs
from soda.contracts.impl.variable_resolver import VariableResolver


class SodaCloud:
    def __init__(
        self,
        host: str,
        api_key_id: str,
        api_key_secret: str,
        token: str | None,
        port: str | None,
        scheme: str,
        logs: Logs,
    ):
        self.host: str = host
        self.api_key_id: str = api_key_id
        self.api_key_secret: str = api_key_secret
        self.token: str | None = token
        self.port: str | None = port
        self.scheme: str = scheme
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs

    @classmethod
    def from_environment_variables(cls) -> SodaCloud:
        return SodaCloud.from_dict({})

    @classmethod
    def from_yaml_file(cls, soda_cloud_yaml_file_path: str) -> SodaCloud:
        """
        soda_cloud: SodaCloud = SodaCloud.from_yaml_file(file_path)

        :param soda_cloud_yaml_file_path: A file path to a YAML file containing the Soda Cloud configuration
            properties. See TODO for more details on the schema of the YAML file

        :return: A Soda Cloud object containing the Soda Cloud connection properties
        :raises SodaCloudException: if SodaCloud cannot be created for any reason
        """
        logs: Logs = Logs()

        try:
            if not isinstance(soda_cloud_yaml_file_path, str):
                logs.error(
                    f"Couldn't create SodaCloud from yaml file. Expected str in parameter "
                    f"soda_cloud_yaml_file_path={soda_cloud_yaml_file_path}, but was '{type(soda_cloud_yaml_file_path)}"
                )
            if not len(soda_cloud_yaml_file_path) > 1:
                logs.error(f"Couldn't create SodaCloud from yaml file. soda_cloud_yaml_file_path is an empty string")

            with open(file=soda_cloud_yaml_file_path) as f:
                soda_cloud_yaml_str = f.read()
                return cls.from_yaml_str(soda_cloud_yaml_str=soda_cloud_yaml_str, logs=logs)
        except Exception as e:
            raise SodaException(f"Couldn't create SodaCloud from yaml file '{soda_cloud_yaml_file_path}'") from e

    @classmethod
    def from_yaml_str(cls, soda_cloud_yaml_str: str, logs: Logs | None = None) -> SodaCloud:
        """
        soda_cloud_yaml_str: str = "...YAML string for Soda Cloud configuration properties..."
        soda_cloud: SodaCloud = SodaCloud.from_yaml_str(soda_cloud_yaml_str) as connection:
        # do stuff with Soda Cloud.

        More docs on SodaCloud: see ../../docs/creating_a_soda_cloud.md
        """

        if not logs:
            logs = Logs()

        if not isinstance(soda_cloud_yaml_str, str):
            logs.error(
                f"Expected a string for parameter soda_cloud_yaml_str, " f"but was '{type(soda_cloud_yaml_str)}'"
            )

        if soda_cloud_yaml_str == "":
            logs.error(f"soda_cloud_yaml_str must be non-emtpy, but was ''")

        variable_resolver = VariableResolver(logs=logs)
        resolved_soda_cloud_yaml_str: str = variable_resolver.resolve(soda_cloud_yaml_str)

        soda_cloud_dict: object | None = None
        try:
            yaml = YAML()
            yaml.preserve_quotes = True
            soda_cloud_dict = yaml.load(resolved_soda_cloud_yaml_str)
        except MarkedYAMLError as e:
            mark = e.context_mark if e.context_mark else e.problem_mark
            line = (mark.line + 1,)
            column = (mark.column + 1,)
            logs.error(f"YAML syntax error: {e} | line={line} | column={column}")

        if not isinstance(soda_cloud_dict, dict):
            logs.error(f"Content of the SodaCloud YAML file must be a YAML object, " f"but was {type(soda_cloud_dict)}")

        return cls.from_dict(soda_cloud_dict=soda_cloud_dict, logs=logs)

    @classmethod
    def from_dict(cls, soda_cloud_dict: dict | None = None, logs: Logs | None = None) -> SodaCloud | None:
        if soda_cloud_dict is None:
            soda_cloud_dict = {}

        if not logs:
            logs = Logs()

        if not isinstance(soda_cloud_dict, dict):
            logs.error(f"soda_cloud_dict must be a dict, but was {type(soda_cloud_dict)}")

        host: str = cls.__get_configuration_value(
            soda_cloud_dict=soda_cloud_dict, key="host", logs=logs, default_value="cloud.soda.io"
        )
        api_key_id: str = cls.__get_configuration_value(
            soda_cloud_dict=soda_cloud_dict, key="api_key_id", logs=logs, required=True
        )
        api_key_secret: str = cls.__get_configuration_value(
            soda_cloud_dict=soda_cloud_dict, key="api_key_secret", logs=logs, required=True
        )
        token: str | None = cls.__get_configuration_value(soda_cloud_dict=soda_cloud_dict, key="token", logs=logs)
        port: str | None = cls.__get_configuration_value(soda_cloud_dict=soda_cloud_dict, key="port", logs=logs)
        scheme: str = cls.__get_configuration_value(
            soda_cloud_dict=soda_cloud_dict, key="scheme", logs=logs, default_value="https"
        )
        if not isinstance(api_key_id, str) or not isinstance(api_key_secret, str):
            return None
        return SodaCloud(
            host=host,
            api_key_id=api_key_id,
            api_key_secret=api_key_secret,
            token=token,
            port=port,
            scheme=scheme,
            logs=logs,
        )

    @classmethod
    def __get_configuration_value(
        cls,
        soda_cloud_dict: dict,
        key: str,
        logs: Logs,
        default_value: str | None = None,
        required: bool = False,
    ) -> str:
        environment_key_lower = f"soda_cloud_{key}".lower()
        if environment_key_lower in os.environ:
            return os.environ[environment_key_lower]
        environment_key_upper = environment_key_lower.upper()
        if environment_key_upper in os.environ:
            return os.environ[environment_key_upper]
        if key in soda_cloud_dict:
            return soda_cloud_dict[key]
        if not required:
            return default_value
        logs.error(f"Key {environment_key_lower} not found in Soda Cloud configuration")



from __future__ import annotations

import logging
from numbers import Number
from typing import Any

from ruamel.yaml import YAML, CommentedMap, CommentedSeq, round_trip_dump
from ruamel.yaml.error import MarkedYAMLError

from soda.contracts.connection import SodaException
from soda.contracts.impl.logs import Logs

logger = logging.getLogger(__name__)


class YamlParser:

    def __init__(self, logs: Logs | None = None):
        self.ruamel_yaml: YAML = YAML()
        self.ruamel_yaml.preserve_quotes = True
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs if logs else Logs()

    def parse_yaml_str(self, yaml_str: str) -> object:
        """
        Parses the YAML string using the ruamel parser.  Returns a data structure representing
        the YAML with dicts (ruamel CommentedMap) and array (ruamel CommentedSeq) amd basic
        python types.
        Swallows Ruamel parsing exceptions and logs them either to the self.logs object
        """
        try:
            return self.ruamel_yaml.load(yaml_str)
        except MarkedYAMLError as e:
            location = self.get_location_from_yaml_error(e)
            raise SodaException(f"YAML syntax error at {location}: {e}") from e

    @classmethod
    def get_location_from_yaml_error(cls, e):
        mark = e.context_mark if e.context_mark else e.problem_mark
        return YamlLocation(
            line=mark.line + 1,
            column=mark.column + 1,
        )


class YamlWriter:
    def __init__(self, logs: Logs | None = None):
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs if logs else Logs()

    def write_to_yaml_str(self, yaml_object: object) -> str:
        try:
            return round_trip_dump(yaml_object)
        except Exception as e:
            self.logs.error(f"Couldn't write SodaCL YAML object: {e}", exception=e)


class YamlWrapper:
    """
    Wraps Ruamel YAML objects into YamlValue's that provide
     - capturing all errors into a central logs object instead of raising an exception on the first problem
     - convenience read_* methods on the YamlObject for writing parsing code
     - a more convenient way to access the line and column information (location)
    """

    def __init__(self, logs: Logs = Logs()):
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs

    def wrap(self, ruamel_value: object) -> YamlValue:
        if isinstance(ruamel_value, str):
            return YamlString(ruamel_value=ruamel_value)
        if isinstance(ruamel_value, bool):
            return YamlBoolean(ruamel_value=ruamel_value)
        if isinstance(ruamel_value, Number):
            return YamlNumber(ruamel_value=ruamel_value)
        if ruamel_value is None:
            return YamlNull()
        if isinstance(ruamel_value, CommentedMap):
            return YamlObject(ruamel_value=ruamel_value, yaml_wrapper=self)
        if isinstance(ruamel_value, CommentedSeq):
            return YamlList(ruamel_value=ruamel_value, yaml_wrapper=self)
        logging.error(f"Unsupported Ruamel YAML object type: {type(ruamel_value).__name__}\n{str(ruamel_value)}")


class YamlLocation:
    def __init__(self, line: int, column: int):
        self.line: int = line
        self.column: int = column

    def __str__(self):
        return f"(line {self.line},col {self.column})"


class YamlValue:
    """
    Base class for yaml data structure objects
    """

    def __init__(self, ruamel_value: object):
        self.ruamel_value: Any = ruamel_value
        self.location: YamlLocation | None = None

    def set_location(self, location: YamlLocation | None) -> None:
        self.location = location

    def unpacked(self) -> Any:
        return self.ruamel_value


class YamlString(YamlValue):
    def __init__(self, ruamel_value: str):
        super().__init__(ruamel_value)


class YamlNumber(YamlValue):
    def __init__(self, ruamel_value: Number):
        super().__init__(ruamel_value)


class YamlBoolean(YamlValue):
    def __init__(self, ruamel_value: bool):
        super().__init__(ruamel_value)


class YamlNull(YamlValue):
    def __init__(self):
        super().__init__(None)


class YamlObject(YamlValue):
    def __init__(self, ruamel_value: CommentedMap, yaml_wrapper: YamlWrapper):
        super().__init__(ruamel_value)
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = yaml_wrapper.logs
        self.yaml_dict: dict[str, YamlValue] = {
            key: self.__convert_map_value(ruamel_object=ruamel_value, key=key, value=value, yaml_wrapper=yaml_wrapper)
            for key, value in ruamel_value.items()
        }

    def __convert_map_value(self, ruamel_object: CommentedMap, key, value, yaml_wrapper: YamlWrapper) -> YamlValue:
        ruamel_location = ruamel_object.lc.value(key)
        line: int = ruamel_location[0]
        column: int = ruamel_location[1]
        yaml_value = yaml_wrapper.wrap(ruamel_value=value)
        yaml_value.set_location(YamlLocation(line=line, column=column))
        return yaml_value

    def unpacked(self) -> Any:
        return {k: v.unpacked() for k, v in self.yaml_dict.items()}

    def __iter__(self):
        return iter(self.yaml_dict)

    def __contains__(self, key):
        return key in self.yaml_dict

    def __len__(self) -> int:
        return len(self.yaml_dict)

    def keys(self):
        return self.yaml_dict.keys()

    def items(self):
        return self.yaml_dict.items()

    def get(self, index) -> YamlValue | None:
        return self.yaml_dict.get(index)

    def read_yaml_object(self, key: str) -> YamlObject | None:
        """
        An appropriate error log is generated if the value is not a YamlObject or if the key is missing
        :return: a YamlObject if the value for the key is a YamlObject, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlObject, required=True, default_value=None)

    def read_yaml_object_opt(self, key: str) -> YamlObject | None:
        """
        An appropriate error log is generated if the value is not a YamlObject
        :return: a YamlObject if the value for the key is a YamlObject, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlObject, required=False, default_value=None)

    def read_yaml_list(self, key: str) -> YamlList | None:
        """
        An appropriate error log is generated if the value is not a YamlList or if the key is missing
        :return: a YamlList if the value for the key is a YamlList, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlList, required=True, default_value=None)

    def read_yaml_list_opt(self, key: str) -> YamlList | None:
        """
        An appropriate error log is generated if the value is not a YamlObject
        :return: a YamlObject if the value for the key is a YamlObject, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlList, required=False, default_value=None)

    def read_yaml_string_opt(self, key: str, default_value: str | None = None) -> YamlString | None:
        """
        An appropriate error log is generated if the value is not a string
        :return: a YamlString if the value for the key is a YAML string, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlString, required=False, default_value=default_value)

    def read_yaml_string(self, key: str) -> YamlString | None:
        """
        An appropriate error log is generated if the value is not a string or if the key is missing.
        :return: a YamlString if the value for the key is a YAML string, otherwise None.
        """
        return self.read_value(key=key, expected_type=YamlString, required=True)

    def read_string_opt(self, key: str, default_value: str | None = None) -> str | None:
        """
        An appropriate error log is generated if the value is not a string
        :return: a str if the value for the key is a YAML string, otherwise None.
        """
        yaml_string_value = self.read_value(
            key=key, expected_type=YamlString, required=False, default_value=default_value
        )
        return yaml_string_value.unpacked() if isinstance(yaml_string_value, YamlString) else None

    def read_string(self, key: str) -> str | None:
        """
        An appropriate error log is generated if the value is not a string or if the key is missing.
        :return: a YamlString if the value for the key is a YAML string, otherwise None.
        """
        yaml_string_value = self.read_value(key=key, expected_type=YamlString, required=True, default_value=None)
        return yaml_string_value.unpacked() if isinstance(yaml_string_value, YamlString) else None

    def read_list_strings(self, key: str) -> list[str] | None:
        yaml_list_value = self.read_value(key=key, expected_type=YamlList, required=True, default_value=None)
        if isinstance(yaml_list_value, YamlList):
            value = yaml_list_value.unpacked()
            if all(isinstance(e, str) for e in value):
                return value
            else:
                self.logs.error(f"Not all elements in list are strings: {yaml_list_value.location}")

    def read_bool(self, key: str) -> bool | None:
        """
        An appropriate error log is generated if the value is not a bool or if the key is missing
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        yaml_boolean_value = self.read_value(key=key, expected_type=YamlBoolean, required=True, default_value=None)
        return yaml_boolean_value.unpacked() if isinstance(yaml_boolean_value, YamlString) else None

    def read_bool_opt(self, key: str, default_value: bool | None = None) -> bool | None:
        """
        An appropriate error log is generated if the value is not a bool.
        :return: a bool if the value for the key is a YAML boolean, otherwise None.
        """
        yaml_boolean_value = self.read_value(
            key=key, expected_type=YamlBoolean, required=False, default_value=default_value
        )
        return yaml_boolean_value.unpacked() if isinstance(yaml_boolean_value, YamlBoolean) else None

    def read_number(self, key: str) -> Number | None:
        """
        An appropriate error log is generated if the value is not a number or if the key is missing
        :return: a bool if the value for the key is a YAML number, otherwise None.
        """
        yaml_number_value = self.read_value(key=key, expected_type=YamlNumber, required=True, default_value=None)
        return yaml_number_value.unpacked() if isinstance(yaml_number_value, YamlNumber) else None

    def read_number_opt(self, key: str, default_value: Number | None = None) -> Number | None:
        """
        An appropriate error log is generated if the value is not a number.
        :return: a Number if the value for the key is a YAML number, otherwise None.
        """
        yaml_number_value = self.read_value(
            key=key, expected_type=YamlNumber, required=False, default_value=default_value
        )
        return yaml_number_value.unpacked() if isinstance(yaml_number_value, YamlNumber) else None

    def read_value(
        self,
        key: str,
        expected_type: type = None,
        required: bool = False,
        default_value=None,
    ) -> YamlValue:
        if key not in self.yaml_dict:
            if required:
                msg = f"'{key}' is required {self.location}"
                if self.logs:
                    self.logs.error(msg)
                else:
                    logger.error(msg)
            return default_value
        yaml_value: YamlValue = self.yaml_dict.get(key)
        if expected_type is not None and not isinstance(yaml_value, expected_type):
            msg = f"'{key}' expected a {expected_type.__name__}, but was {type(yaml_value).__name__} {yaml_value.location}"
            if self.logs:
                self.logs.error(msg)
            else:
                logger.error(msg)
        return yaml_value


class YamlList(YamlValue):
    def __init__(self, ruamel_value: CommentedSeq, yaml_wrapper: YamlWrapper):
        super().__init__(ruamel_value)
        self.value: list[YamlValue] = [
            self.__convert_array_value(
                ruamel_value=ruamel_list_value, commented_seq=ruamel_value, index=index, yaml_wrapper=yaml_wrapper
            )
            for index, ruamel_list_value in enumerate(ruamel_value)
        ]

    def __convert_array_value(
        self, ruamel_value, commented_seq: CommentedSeq, index: int, yaml_wrapper: YamlWrapper
    ) -> YamlValue:
        ruamel_location = commented_seq.lc.key(index)
        line: int = ruamel_location[0]
        column: int = ruamel_location[1]
        yaml_value = yaml_wrapper.wrap(ruamel_value=ruamel_value)
        yaml_value.set_location(YamlLocation(line=line, column=column))
        return yaml_value

    def unpacked(self) -> Any:
        return [e.unpacked() for e in self.value]

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, key):
        return key in self.value

    def __len__(self) -> int:
        return len(self.value)

    def read_yaml_objects(self) -> list[YamlObject]:
        return [e for e in self.value if isinstance(e, YamlObject)]



from __future__ import annotations

import logging
from typing import Dict

import soda.common.logs as soda_common_logs
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError
from soda.execution.data_source import DataSource

from soda.contracts.impl.logs import Logs
from soda.contracts.impl.variable_resolver import VariableResolver

logger = logging.getLogger(__name__)


class SodaException(Exception):
    """
    See also adr/03_exceptions_vs_error_logs.md
    """

    def __init__(self, message: str | None = None, contract_result: ContractResult | None = None):
        from soda.contracts.contract import ContractResult

        self.contract_result: ContractResult = contract_result
        message_parts: list[str] = []
        if message:
            message_parts.append(message)
        if self.contract_result:
            message_parts.append(str(self.contract_result))
        exception_message: str = "\n".join(message_parts)
        super().__init__(exception_message)


class Connection:
    """
    A wrapper for DBAPI a connection to handle all database differences. Usage:

    with Connection.from_dict({
      "type": "postgres",
      "host": "localhost",
      "database": "soda",
      "user": "postgres",
      "password": "<PASSWORD>"
    }) as connection:
        # Do stuff with connection

    When creating a connection from YAML, you can use ${VAR} to resolve environment variables.
    Recommended for credentials.

    The 'type' property is used to determine the type of connection and driver.
    All the other properties are passed in the DBAPI connect method to create the DBAPI connection.
    """

    def __init__(self, dbapi_connection: object | None = None, name: str | None = None, logs: Logs | None = None):
        self.dbapi_connection = dbapi_connection
        self.name: str | None = name
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs if logs else Logs()

    @classmethod
    def from_yaml_file(cls, connection_yaml_file_path: str) -> Connection:
        """
        with Connection.from_yaml_file(file_path) as connection:
            # Do stuff with connection

        Use ${YOUR_PASSWORD} to resolve environment variables. Recommended for credentials.

        :param connection_yaml_file_path: A file path to a YAML file containing the connection configuration properties.
        :return: an open connection, if no exception is raised
        :raises SodaConnectionException: if the connection cannot be established for any reason
        """

        logs: Logs = Logs()
        try:
            if not isinstance(connection_yaml_file_path, str):
                logs.error(
                    f"Couldn't create connection from yaml file. Expected str in parameter "
                    f"connection_yaml_file_path={connection_yaml_file_path}, but was '{type(connection_yaml_file_path)}"
                )
            elif len(connection_yaml_file_path) <= 1:
                logs.error("Couldn't create connection from yaml file. connection_yaml_file_path is an empty string")
            else:
                with open(file=connection_yaml_file_path) as f:
                    connection_yaml_str = f.read()
                    return cls.from_yaml_str(connection_yaml_str=connection_yaml_str, logs=logs)
        except Exception as e:
            logs.error(f"Couldn't create connection from yaml file '{connection_yaml_file_path}': {e}", exception=e)
        return Connection(logs=logs)

    @classmethod
    def from_yaml_str(
        cls, connection_yaml_str: str, variables: Dict[str, str] | None = None, logs: Logs | None = None
    ) -> Connection:
        """
        # TODO specify where the connection configuration properties are being documented
        connection_yaml_str: str = "...YAML string for connection configuration properties..."
        with Connection.from_yaml_str(connection_yaml_str) as connection:
           # do stuff with connection

        Use ${YOUR_PASSWORD} to resolve environment variables. Recommended for credentials. Resolving will
        first use the variables parameter and otherwise fall back to environment variables.  A SodaConnectionException
        will be raised if a variable is used and cannot be resolved.

        :param connection_yaml_str: A YAML string containing the connection configuration properties.
        :param variables: Optional dictionary of variables that will be used to resolve variables before checking
        environment variables.
        :return: an open connection, if no exception is raised
        :raises SodaConnectionException: if the connection cannot be established for any reason
        """

        if not logs:
            logs = Logs()

        if not isinstance(connection_yaml_str, str):
            logs.error(f"Expected a string for parameter connection_yaml_str, but was '{type(connection_yaml_str)}'")

        if connection_yaml_str == "":
            logs.error("connection_yaml_str must be non-emtpy, but was ''")

        try:
            variable_resolver = VariableResolver(variables=variables, logs=logs)
            resolved_connection_yaml_str = variable_resolver.resolve(connection_yaml_str)
        except BaseException as e:
            logs.error(f"Could not resolve variables in connection YAML: {e}", exception=e)

        try:
            yaml = YAML()
            yaml.preserve_quotes = True
            connection_dict = yaml.load(resolved_connection_yaml_str)
            if not isinstance(connection_dict, dict):
                logs.error(
                    f"Content of the connection YAML file must be a YAML object, but was {type(connection_dict)}"
                )
            return cls.from_dict(connection_dict)
        except MarkedYAMLError as e:
            mark = e.context_mark if e.context_mark else e.problem_mark
            line = (mark.line + 1,)
            column = (mark.column + 1,)
            logs.error(f"YAML syntax error: {e} | line={line} | column={column}")

    @classmethod
    def from_dict(cls, connection_dict: dict, logs: Logs | None = None) -> Connection:
        """
        with Connection.from_dict({
          "type": "postgres",
          "host": "localhost",
          "database": "soda",
          "user": "postgres",
          "password": "<PASSWORD>"
        }) as connection:
            # Do stuff with connection

        :return: an open connection, if no exception is raised
        :raises SodaConnectionException: if the connection cannot be established for any reason
        """
        if not logs:
            logs = Logs()

        connection_type: str | None = None

        if not isinstance(connection_dict, dict):
            logs.error(f"connect_properties must be a object, but was {type(connection_dict)}")
        elif "type" not in connection_dict:
            logs.error("'type' is required, but was not provided")
        else:
            connection_type = connection_dict.get("type")
            if not isinstance(connection_type, str):
                logs.error(f"'type' must be a string, but was  {type(connection_type)}")
        return DataSourceConnection(connection_type=connection_type, connection_dict=connection_dict, logs=logs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except Exception as e:
            logger.warning(f"Could not close connection: {e}")

    def close(self) -> None:
        """
        Closes te connection. This method will not throw any exceptions.
        Check errors with has_errors or assert_no_errors.
        """
        if self.dbapi_connection:
            try:
                self.dbapi_connection.close()
            except Exception as e:
                logger.warning(f"Could not close the dbapi connection: {e}")


class DataSourceConnection(Connection):

    def __init__(self, connection_type: str, connection_dict: dict, logs: Logs | None = None):
        # consider translating postgres schema search_path option
        # options = f"-c search_path={schema}" if schema else None
        try:
            self.data_source = DataSource.create(
                logs=soda_common_logs.Logs(logger=logger),
                data_source_name=f"{connection_type}_ds",
                data_source_type=connection_type,
                data_source_properties=connection_dict,
            )
            self.data_source.connect()
        except Exception as e:
            logs.error(message=f"Could not create the connection: {e}", exception=e)
        name: str | None = connection_dict.get("name", connection_type)
        super().__init__(dbapi_connection=self.data_source.connection, name=name, logs=logs)
